print("import: 'locust'")
import locust

print("import: 'locust.contrib'")
import locust.contrib

print("import: 'locust.rpc'")
import locust.rpc

print("import: 'locust.test'")
import locust.test

print("import: 'locust.util'")
import locust.util

